import sys
import os.path
import json
sys.path.insert(0,os.path.abspath(os.path.join( os.path.dirname(__file__), "../"))) # Use local grew lib

from grewpy import Graph

g_str = """
{
  "nodes": {
    "A": "A",
    "B": "B",
    "C": "C"
  },
  "edges": [
    { "src": "A", "label": "X", "tar": "B"},
    { "src": "B", "label": "Y", "tar": "C"},
    { "src": "C", "label": "Z", "tar": "A"}
  ],
  "order": [ "A", "B" ]
}
"""

print ("---- build a graph from a string ----")
g1 = Graph(g_str)

print (f"|nodes| = {(len (g1))}")

print (g1["A"])
print ("---- sucs 2 ----")
print (g1.sucs["A"])
print ("---- sucs 1 ----")
print (g1.sucs["B"])


print("---modify a graph---")
print("----- add an edge ---")
g1.sucs['C'].append(("A",{"1":"E"}))
print(g1.sucs['C'])
print("---remove C edges---")
g1.sucs['C'] = []
#g1.sucs['D'].append(('C',{'1':'F'}))



print ("---- JSON output of a graph ----")
print (json.dumps(g1.json_data(), indent=4))



import sys,os

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../")))  # Use local grew lib

from grewpy import Corpus, GRSDraft, Rule, Request, Commands, GRS, set_config
from grewpy import Grew_web
set_config("sud")

corpus = Corpus("examples/resources/pud_10.conllu")

grs_text = """
rule r { pattern { N[upos="NOUN"] } commands { N.upos=N } }
strat main { Onf (r) }
"""
grs = GRS(grs_text)

web = Grew_web()
web.load_corpus(Corpus(corpus[0:3])) #to get a few samples in the corpus
web.load_grs(grs)

print(web.url())

# TEST 1: start Grew-web with a GRS and a corpus
#grs.grew_web(corpus)

# TEST 2: start Grew-web with a GRS
# grs.grew_web()

# TEST 3: start Grew-web with a corpus
# corpus.grew_web()




''' Utility tools to connect to ocaml GREW'''

from subprocess import Popen, PIPE
import time
import socket
import os.path
import json
import os
import sys

from .grew import GrewError

host = 'localhost'
port = 8888
remote_ip = ''
caml_pid = None

request_counter = 0 #number of request to caml

import signal
def preexec_function ():
    signal.signal(signal.SIGINT, signal.SIG_IGN)

def pid_exist(pid):
    try:
        os.kill(pid,0)
        return 1
    except:
        return 0

def init():
    global port, remote_ip, caml_pid
    grewpy = "grewpy_backend"
    if not pid_exist(caml_pid):
        python_pid = os.getpid()
        caml = Popen(
            [grewpy, "--caller", str(python_pid)],
            preexec_fn=preexec_function,
            stdout=PIPE
        )
        port = int(caml.stdout.readline().strip())
        caml_pid = caml.pid
        #wait for grew's lib answer
        time.sleep(0.1)
        if caml.poll() == None:
            print ("connected to port: " + str(port), file=sys.stderr)
            remote_ip = socket.gethostbyname(host)
            return (caml)
        else:
            print ("Failed to connect", file=sys.stderr)
            exit (1)

def connect():
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((remote_ip, port))
        return s
    except socket.error:
        raise GrewError('Failed to create socket. Make sure that you have called grew.init.')
    except socket.gaierror:
        print('[GREW] Hostname could not be resolved. Sorry\n', file=sys.stderr)


packet_size=32768

def send_and_receive(msg):
    global request_counter
    try:
        request_counter += 1
        stocaml = connect()
        json_msg = json.dumps(msg).encode(encoding='UTF-8')
        len_string = "%010d" % len(json_msg)
        stocaml.sendall(len_string.encode(encoding='UTF-8'))

        packet_nb = len(json_msg) // packet_size
        for i in range (packet_nb):
            stocaml.sendall(json_msg[packet_size*i:packet_size*(i+1)])
        stocaml.sendall(json_msg[packet_nb*packet_size:])
        camltos = bytes()
        reply_len = int(stocaml.recv(10))

        camltos = b''
        while len(camltos) < reply_len:
            packet = stocaml.recv(reply_len - len(camltos))
            if not packet:
                return None
            camltos += packet
        stocaml.close()

        reply = json.loads(camltos.decode(encoding='UTF-8'))
        if reply["status"] == "OK":
            try:
                return reply["data"]
            except:
                return None
        elif reply["status"] == "ERROR":
            raise GrewError({"function": msg["command"], "message": reply["message"]})
    except socket.error:
        raise GrewError({"function": msg["command"], "message" : 'Socket error'})
    except AttributeError as e: # connect issue
        raise GrewError({"function": msg["command"], "message" : e.value})




import sys, os, json

sys.path.insert(0, os.path.abspath(os.path.join( os.path.dirname(__file__), "../"))) # Use local grew lib

from grewpy import Graph, CorpusDraft, Request, Corpus, request_counter

pud_file = "examples/resources/pud_10.conllu"
pud = Corpus(pud_file)

print (pud.edge_diff(pud))
exit (0)
print ("\n=============== len ===============")
print(f"nb of graph in {pud_file} = {len(pud)}")
print (request_counter())

print ("\n=============== Get one graph ===============")
sent_id="n01003013"
graph = pud[sent_id]
print(f"nb of nodes of {sent_id} = ", len(graph))
print (request_counter())

print(f"len(pud[0]) = {len(pud[0])}")
print(f"len(pud[-1]) = {len(pud[-1])}")
print(f"[len(g) for g in pud[-3:]] = {[len(g) for g in pud[-3:]]}")
#other forms pud[-3:-1], pud[1:7:2], ...


print ("\n=============== Iteration on graphs of a corpus ===============")
print ("⚠️  generate one request to Grew backend for each graph")
acc = 0
for sent_id in pud.get_sent_ids():
  acc += len(pud[sent_id])
print(f"nb of nodes in {pud_file} = ", acc)
print (request_counter())

print ("\n=============== Count request in a corpus ===============")
upos="ADV"
req1 = Request(f"X[upos={upos}]; Y -[advmod]-> X")
req2 = Request(f"X[upos={upos}]", "Y -[advmod]-> X")
req3 = Request.parse(f"pattern {{ X[upos={upos}]; Y -[advmod]-> X }}")
print(" ----- basic count -----")
print(f"nb of {upos} in {pud_file} = ", pud.count(req1))
print(f"nb of {upos} in {pud_file} = ", pud.count(req2))
print(f"nb of {upos} in {pud_file} = ", pud.count(req3))


exit(0)

print (" ----- count with clustering -----")
print(f"nb of {upos} in {pud_file}, clustered by lemma:")
print (json.dumps(pud.count(req, ["X.lemma"]), indent=2))
print (request_counter())



corpus = CorpusDraft(pud)
print("\n=============== Iteration on graphs of a corpus ===============")
acc = 0
for sent_id in corpus:
  acc += len(corpus[sent_id])
print(f"nb of nodes in {pud_file} = ", acc)
print (request_counter())

def clear_edges(graph):
    for n in graph:
        graph.sucs[n] = []

for sid in corpus:
  clear_edges(corpus[sid])

noedge_corpus = Corpus(corpus)
print(" ----- counting nsubj within corpus -----")
dep = "nsubj"
req = Request(f"X[];Y[];X -[{dep}]-> Y")
print(f"nb of {dep} in {pud_file} = ", pud.count(req))
print(f"nb of {dep} in noedge_corpus = ", noedge_corpus.count(req))

exit(0)



req = Request(f"X -[fixed]-> Y")
print (" ----- count with clustering -----")
print(f"nb of fixed in {pud_file}, clustered by head.lemma:")
print (json.dumps(pud.count(req, ["X.lemma"]), indent=2))

g = pud[sent_id]
for n in g:
  for (s,e) in g.suc(n):
    if e  == {'1' : 'fixed'}:
      print(f"\n{n} : {g[n]} -[{e}]-> {s} : {g[s]}\n")

clear_edges(g)
#pud[sent_id] = g
# NOTE: the next line does not work properly (not __set_item__ called), have a look to https://stackoverflow.com/questions/26189090/how-to-detect-if-any-element-in-a-dictionary-changes
# clear_edges(pud[sent_id]) ==> WARNING: does not change pud!




import classifier
import numpy as np
import re

# UU$se local grew lib
import os
import sys
sys.path.insert(0, os.path.abspath(
    os.path.join(os.path.dirname(__file__), "../")))
from grewpy.graph import Fs_edge
from grewpy.sketch import Sketch
from grewpy import Corpus, GRS
from grewpy import Request, Rule, Commands, Add_edge, GRSDraft, CorpusDraft, Package, Delete_feature

cpt = iter(range(1000000))

def module_name(*t):
    x = str(t)
    x = re.sub(r"\W","",x)
    return x

class WorkingGRS(GRSDraft):
    def __init__(self, *args, **kwargs):
        """
        local class: like a GRSDraft with an additional evaluation of the rules
        each rule is evaluated by two numbers: number of good application, 
        total number of application
        """
        super().__init__(*args, **kwargs)
        self.eval = {x: (0, 0) for x in self}

    def __setitem__(self, __key, __value):
        self.eval[__key] = __value[1]
        return super().__setitem__(__key, __value[0])

    def __ior__(self, other):
        super().__ior__(other)
        self.eval |= other.eval
        return self

def anomaly(obsL,  threshold):
    """
        L is a key within self
        return for L an edge and its occurrence evaluation 
        and number of total occurrences if beyond base_threshold
    """
    s = sum(obsL.values())
    for x, v in obsL.items():
        if v > threshold * s and x:
            return (x, v, s)
    return None, None, None

def build_rules(sketch, observation, param, rule_name, verbose=False):
    """
    search a rule adding an edge X -> Y, given a sketch  
    we build the clusters, then
    for each pair (X, upos=U1), (Y, upos=U2), we search for 
    some edge e occuring at least with probability base_threshold
    in which case, we define a rule R: 
    base_pattern /\ [X.upos=U1] /\ [Y.upos=U2] => add_edge X-[e]-Y
    """
    def crit_to_request(crit, val):
        if ".label" in crit:
            edge_name = re.match("(.*?).label", crit).group(1)
            clauses = Fs_edge.decompose_edge(val)
            return ";".join((f'{edge_name}.{a}="{b}"' for a, b in clauses.items()))
        return f'{crit}="{val}"'
    rules = WorkingGRS()
    loose_rules = WorkingGRS()
    for parameter in observation:
        if verbose:
            print(observation[parameter])
        x, v, s = anomaly(observation[parameter], param["base_threshold"])
        if x:
            extra_pattern = [crit_to_request(crit, val) for (
                crit, val) in zip(sketch.cluster_criterion, parameter)]
            P = Request(sketch.P, *extra_pattern)
            x0 = Fs_edge(x)  # {k:v for k,v in x}
            c = Add_edge("X", x0, "Y")
            R = Rule(P, Commands(c))
            rn = module_name(f"_{'_'.join(parameter)}_{rule_name}")
            rules[rn] = (R, (x, (v, s)))
        else:
            ...
    return rules, loose_rules

def refine_rule(R, corpus, param) -> list[Rule]:
    """
    Takes a request R, tries to find variants
    the result is the list of rules that refine pattern R
    for DEBUG, we return the decision tree classifier
    """
    res = []
    matchings = corpus.search(R)
    clf = classifier.Classifier(matchings, corpus, param)
    if clf.clf:
        branches = clf.find_classes(param)  # extract interesting branches
        for node in branches:
            branch = branches[node]
            request = Request(R)  # builds a new Request
            for feature_index, negative in branch:
                n, feat, feat_value = clf.fpat[feature_index]
                feat_value = feat_value.replace('"', '\\"')
                if negative:
                    request = request.without(f'{n}[{feat}="{feat_value}"]')
                else:
                    request.append("pattern", f'{n}[{feat}="{feat_value}"]')
            e = clf.y1[clf.clf.tree_.value[node].argmax()]
            if e:  # here, e == None if there is no edges X -> Y
                rule = Rule(request, Commands(Add_edge("X", e, "Y")))
                res.append(rule)
    return res, clf


def refine_rules(Rs, corpus, param, verbose=False):
    """
    as above, but applies on a list of rules
    and filter only "correct" rules, see `param`
    return the list of refined version of rules Rs
    """
    Rse = WorkingGRS()
    for rule_name in Rs.rules():
        R = Rs[rule_name]
        v, s = Rs.eval[rule_name][1]
        if v < param["min_occurrence_nb"]:
            pass
        elif v/s < param["valid_threshold"]:
            new_r, clf = refine_rule(R.request, corpus, param)
            if len(new_r) >= 1:
                if verbose:
                    print("--------------------------replace")
                    print(R)
                for r in new_r:
                    if verbose:
                        print("by : ")
                        print(r)
                    X = ",".join(f'{k}="{v}"' for k,
                                 v in r.commands[0].e.items())
                    x1 = corpus.count(Request(r.request, f"X-[{X}]->Y"))
                    x2 = corpus.count(
                        Request(r.request).without(f"X-[{X}]->Y"))
                    Rse[f"{rule_name}_enhanced{next(cpt)}"] = (
                        r, x1/(x1+x2+1e-10))
        else:
            Rse[rule_name] = (R, s)
    return Rse


"""
Learning sketches
"""
def edge_between_X_and_Y(P):
    return Request(P, 'e:X->Y')


def no_edge_between_X_and_Y(P):
    return Request(P).without('e:X-[^ANCESTOR]->Y')


def simple_sketch(r):
    return Sketch(r, ["X.upos", "Y.upos"], edge_between_X_and_Y, no_edge_between_X_and_Y, "e.label")


def apply_sketches(sketches, corpus, param):
    """
    find rules from sketches
    """
    rules = WorkingGRS()
    loose_rules = WorkingGRS()
    for sketch_name in sketches:
        sketch = sketches[sketch_name]
        r1, l1 = build_rules(sketch, sketch.cluster(
            corpus), param, sketch_name)
        rules |= r1
        loose_rules |= l1
    return rules, loose_rules


def adjacent_rules(corpus: Corpus, param) -> WorkingGRS:
    """
    build all adjacent rules. They are supposed to connect words at distance 1
    """
    sadj = dict()
    sadj["adjacent_lr"] = simple_sketch(Request("X[];Y[head];X<Y"))
    sadj["adjacent_rl"] = simple_sketch(Request("X[];Y[head];Y<X"))
    sadj["adj2_lr"] = sketch_with_parameter(Request("X[];Y[head];Z[];X<Z;Z<Y"),["Z.upos"])
    sadj["adj2_rl"] = sketch_with_parameter(Request("X[];Y[head];Z[];Y<Z;Z<X"),["Z.upos"])
    sadj["no_intermediate_1"] = simple_sketch(
        Request("X[];Y[head];X<<Y").without("Z[];X<<Z;Z<<Y;X.upos=Z.upos"))
    sadj["no_intermediate_2"] = simple_sketch(
        Request("X[];Y[head];X<<Y").without("Z[];X<<Z;Z<<Y;Y.upos=Z.upos"))
    sadj["no_intermediate_3"] = simple_sketch(
        Request("X[];Y[head];Y<<X").without("Z[];Y<<Z;Z<<X;Y.upos=Z.upos"))
    sadj["no_intermediate_4"] = simple_sketch(
        Request("X[];Y[head];Y<<X").without("Z[];Y<<Z;Z<<X;X.upos=Z.upos"))

    nodes = ['f:X -> Z', 'f:Y -> Z', 'f:Z->X', 'f:Z->Y']
    ordres = ['X<Y', 'X>Y', 'Z<Y', 'Z>Y', 'X<Z',
              'X>Z', 'Z<<Y', 'Z>>Y', 'X<<Z', 'X>>Z']
    on_label = [("Z.upos",), ("f.label",), tuple()]
    for ns in nodes:
        for o in ordres:
            for extra in on_label:
                sadj[module_name((ns, o, extra, next(cpt)))] =\
                    Sketch(Request('X[];Y[head]', ns, o), ["X.upos", "Y.upos"] + list(extra),
                           edge_between_X_and_Y,
                           no_edge_between_X_and_Y, "e.label")
    return apply_sketches(sadj, corpus, param)


def sketch_with_parameter(req, extra_labels=[]):
    return Sketch(Request(req), ["X.upos", "Y.upos"] + extra_labels, edge_between_X_and_Y, no_edge_between_X_and_Y, "e.label")


def local_rules(corpus: Corpus, param) -> WorkingGRS:
    local = ["h:U$->V$;U$<<X;U$<<Y;X<<V$;Y<<V$", "h:V$->U$;U$<<X;U$<<Y;X<<V$;Y<<V$"]
    sadj = dict()
    for loc in local:
        sadj[module_name("loc_lr", loc)] = sketch_with_parameter(
            Request(loc, "X[];Y[head];X<Y"))
        sadj[module_name("loc_rl", loc)] = sketch_with_parameter(
            Request(loc, "X[];Y[head];Y<X"))
        sadj[module_name("int_1", loc)] = sketch_with_parameter(Request(
            loc, "X[];Y[head];X<<Y").without("Z[];X<<Z;Z<<Y;X.upos=Z.upos"))
        sadj[module_name("int_2", loc)] = sketch_with_parameter(Request(
            loc, "X[];Y[head];X<<Y").without("Z[];X<<Z;Z<<Y;Y.upos=Z.upos"))
        sadj[module_name("int_3", loc)] = sketch_with_parameter(
            Request(loc, "X[];Y[head];Y<<X").without("Z[];Y<<Z;Z<<X;Y.upos=Z.upos"))
        sadj[module_name("int_4", loc)] = sketch_with_parameter(
            Request(loc, "X[];Y[head];Y<<X").without("Z[];Y<<Z;Z<<X;X.upos=Z.upos"))

    nodes = ['f:X -> Z$', 'f:Y -> Z$', 'f:Z$->X', 'f:Z$->Y']
    ordres = ['X<Y', 'X>Y', 'X<<Y','Y<<X']
    on_label = [("f.label",), tuple()]
    for loc in local:
        for ns in nodes:
            for o in ordres:
                for extra in on_label:
                    sadj[module_name((loc, ns, o, extra, next(cpt)))] =\
                Sketch(Request(loc,'X[];Y[head];e<>h', ns, o), ["X.upos", "Y.upos", "Z$.upos"] + list(extra), 
                edge_between_X_and_Y, 
                no_edge_between_X_and_Y, "e.label")
    return apply_sketches(sadj, corpus, param)



"""
Grew module: anything you want to talk about graphs
"""
import os.path
import re
import copy
import tempfile
import json
import numpy as np

from grewpy.grew import GrewError
from grewpy.network import send_and_receive
from grewpy import utils
from . import network

''' interfaces'''
class Fs_edge(dict):
    def __init__(self,data):
        if isinstance(data,str):
            #try to split a dict
            if "=" in data and "," in data:
                #suppose it is a dictionary
                super().__init__(Fs_edge.decompose_edge(data))
            else:
                super().__init__(Fs_edge.decompose_edge(data))
        elif isinstance(data, dict):
            clauses = dict()
            for k in data:
                Fs_edge.extract(data[k],clauses, k)
            super().__init__(clauses)
        else:
            raise ValueError(f"data is not a feature structure {data}")

    def __eq__(self, other):
        for k,v in self.items():
            if k not in other or v != other[k]:
                return False
        return len(self) == len(other)

    def __neq__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(tuple((sorted(self.items()))))
    
    @staticmethod
    def extract(u, clauses, key='1'):
            if '@' in u:
                u, t = u.split('@')
                clauses['deep'] = t
            if ':' in u:
                u,t = u.split(':')
                clauses['1'] = u
                clauses['2'] = t
            else:
                clauses[key] = u
    @staticmethod
    def decompose_edge(s):
        clauses = dict()
        for it in s.split(","):
            if '=' in s:
                a,b = it.split("=")
                clauses[a] = b
            else:
                Fs_edge.extract(s,clauses)
        return clauses

class Graph():
    """
    a dict mapping node keys to feature structure

    with extra data:
        - an extra dict `sucs` mapping node keys to successors (pair of edge feature,node key)
        - the list `order` containing nodes linearly ordered
        - and the `meta`(data) as a dict

    Param data: either
        - None: return an empty graph
        - a json formatted string
        - a file name containing a json/conll
        - a Graph: return a copy of the graph
        - or named arguments: `features`, `sucs`, `meta` and `order`
    """
    def __init__(self,data=None, **kwargs):

        if isinstance(data, Graph):
            self.features = dict(data.features)
            self._sucs = dict(data._sucs)
            self.meta = dict(data.meta)
            self.order = list(data.order)  
        elif data is None:
            self.features = kwargs.get("features", dict())
            self.order = kwargs.get("order", [])
            self.meta = kwargs.get("meta", dict())
            self._sucs = kwargs.get("sucs", dict())
        elif isinstance(data, dict):
            (features, sucs, meta, order) = Graph._from_json(data)
            self.features = features
            self.order = order
            self.meta = meta            
            self._sucs = sucs
        elif isinstance(data,str):
            #either filename, json or conll
            if os.path.isfile(data):
                req = {"command": "graph_load", "file": data}
                data_json = network.send_and_receive(req)
            else:
                try:
                    data_json = json.loads(data)
                except json.decoder.JSONDecodeError:
                    with tempfile.NamedTemporaryFile(mode="w", delete=True, suffix=".conll") as f:
                        f.write(data)
                        f.flush()  # to be read by others
                        req = {"command": "graph_load", "file": f.name}
                        data_json = network.send_and_receive(req)
            (self.features, self.sucs, self.meta, self.order) = Graph._from_json(data_json)
        else:
            raise GrewError(f"Cannot build Graph with data of type {type(data)}")
        assert isinstance(self.features,dict)

    @staticmethod
    def _from_json(data_json):
        features = data_json["nodes"]
        sucs = dict()
        for edge in data_json.get("edges", []):
            # TODO gestion des "label" implicite
            utils.map_append(sucs, edge["src"],
                             (edge["tar"], Fs_edge(edge["label"])))
        meta = data_json.get("meta", dict())
        order = data_json.get("order", list())
        return (features, sucs, meta, order)

    @classmethod
    def from_json(cls,data_json):
        (features, sucs, meta, order) = Graph._from_json(data_json)
        return cls(features=features, sucs=sucs, order=order, meta=meta)

    def __len__(self):
        """
        return the number of nodes in self
        """
        return len(self.features)

    def __getitem__(self, nid):
        """
        return feature structure corresponding to nid
        """
        return (self.features[nid])

    def __iter__(self):
        return iter(self.features)

    def _gsucs(self):
        return self._sucs

    def _ssucs(self, v):
        self._sucs = v

    def _dsucs(self):
        self._sucs.clear()

    sucs = property(_gsucs, _ssucs, _dsucs, "successor relation")

    def to_dot(self): # TODO fix it
        """
        return a string in dot/graphviz format
        """
        s = 'digraph G{\n'
        for n,fs in self.features.items():
            s += f'{n}[label="'
            label = ["%s:%s" % (f,v.replace('"','\\"')) for f, v in fs.items()]
            s += ",".join(label)
            s += '"];\n'
        s += "\n".join([f'{n} -> {m}[label="{e}"];' for n,suc in self._sucs.items() for e,m in suc])
        return s + '\n}'

    def json_data(self):
        nds = {c:self[c] for c in self.features}
        edg_list = []
        for n in self._sucs:
            for (e,s) in self._sucs[n]:
                if len(s.keys()) == 1 and '1' in s.keys():
                    s = s["1"]
                edg_list.append({"src":f"{n}", "label":s,"tar":f"{e}"})
        return {"nodes" : nds, "edges" : edg_list, "order": self.order, "meta" : self.meta }

    def __str__(self):
        return f"({str(self.features)}, {str(self._sucs)})" # TODO order, meta

    def to_conll(self):
        """
        return a CoNLL string for the given graph
        """
        data = self.json_data()
        req = {"command": "graph_to_conll", "graph": data}
        reply = send_and_receive(req)
        return reply

    def triples(self):
        """
        return the list of edges presented as triples (n,e,s) with n-[e]-> s         
        """
        return list((n, e, s) for n in self._sucs for s,e in self._sucs[n])

    def from_triples(self, triples):
        for n in self:
            self._sucs[n] = []
        for (n,e,m) in triples:
            self._sucs[n].append((e,m))

    def edge(self, n, m):
        """
        given node n and m
        return the "first" label of an edge between n and m if it exists
        """
        if n in self._sucs:
            for (k,v) in self._sucs[n]:
                if k == m:
                    return v
        return None

    def edge_up_to(self, n, m, criterion):
        if n in self._sucs:
            for k,v in self._sucs[n]:
                if k == m and criterion(v):
                    return v

    def edges(self, n, m):
        """
        given node n and m, 
        return the set of edges between n and m
        """
        if n not in self._sucs:
            return []
        return [v for (k,v) in self._sucs[n] if k == m]

    def edges_up_to(self, n, m, criterion):
        """
        search for edges between n and m verifying some criterion
        """
        return [v for (k, v) in self._sucs[n] if k == m and criterion(v)]


    def run(self, Grs, strat="main"):
        return Grs.run(self, strat)

    def apply(self, Grs, strat="main"):
        return Grs.apply(self, strat)

    def edge_diff(self, other, edge_criterion=lambda e: True) -> np.array:
        """
        edge difference between two graphs
        """
        E1 = {(m, repr(e), n) for (m,e,n) in self.triples()  if edge_criterion(e)}  # set of edges as triples
        E2 = {(m, repr(e), n) for (m,e,n) in other.triples() if edge_criterion(e)}  # set of edges as triples
        return np.array([len(E1 & E2), len(E1 - E2), len(E2 - E1)])

    def lower(self, n, m):
        """
        given node n and m in g:
        return True if n < m in g
        """
        if n in self.order and m in self.order:
            return self.order.index(n) < self.order.index(m)
        return False

    def greater(self, n, s):
        """
        return True if n > s in g
        """
        if n in self.order and s in self.order:
            return self.order.index(n) > self.order.index(s)
        return False

    def edge_diff_up_to(self, other, edge_transform=lambda e:e):
        E1 = set()
        for m, e, n in self.triples():
            et = edge_transform(e)
            if et:
                E1.add((m,et,n))
        E2 = set()
        for m, e, n in other.triples():
            et = edge_transform(e)
            if et:
                E2.add((m, et, n))
        return np.array([len(E1 & E2), len(E1 - E2), len(E2 - E1)])
        



import os, sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../"))) 

import grewpy
from grewpy import Corpus, CorpusDraft, Request, grew_web
from grewpy.grew_web import Grew_web

def add_span(g):
    left, right = {i: i for i in g}, {i: i for i in g}
    todo = [i for i in g]
    g._sucs = {i : g._sucs.get(i,[]) for i in g}
    while todo:
        n = todo.pop(0)
        for s, _ in g.sucs[n]:
            if g.lower(left[s], left[n]):
                left[n] = left[s]
                todo.append(n)
            if g.greater(right[s], right[n]):
                right[n] = right[s]
                todo.append(n)
    for i in g:
        g.sucs[i] += [(left[i], {'1': 'LEFT_SPAN'}),
                      (right[i], {'1': 'RIGHT_SPAN'})]
    return g


#corpus = Corpus("resources/fr_pud-sud-test.conllu")
corpus = Corpus("de_gsd-sud-test.conllu")
#corpus = Corpus(c.apply(add_span))  # span

req_root_verb = Request("V[upos=VERB|AUX];R-[root]->V")
res = corpus.search(req_root_verb)

corpus_filtered = Corpus({item["sent_id"] : corpus[item["sent_id"]] for item in res})
print(len(corpus_filtered))
pos2 = Request("R-[root]->V;V[upos=VERB|AUX];V-[^punct|discourse|cc]->L;L<<V")\
    .without("V-[^punct|discourse|cc]->L2;L2<<V").without(
    "V->T;T->L;V<<T;L<<V"
    )
pos3 = Request("R-[root]->V;V[upos=VERB|AUX];V-[^punct|discourse|cc]->L1;L1<<V;V-[^punct|discourse|cc]->L2;L2<<L1")\
    .without("V-[^punct|discourse|cc]->L3;L2<<L3;L3<<V")
pos1 = Request("R-[root]->V;V[upos=VERB|AUX]").without("V-[^punct|discourse|cc]->L;L<<V").without(
    "V[Mood=Imp]")
ppos1 = Request("R-[root]->V;V[upos=VERB|AUX]").without(
    "V-[^punct|discourse|cc]->L;L<<V").without(
    "V -> T;T -> L;V << T;L << V")
gauche0 = corpus_filtered.count(pos1)
ggauche0 = corpus_filtered.count(ppos1)
gauche1 = corpus_filtered.count(pos2)
gauche2p = corpus_filtered.search(pos3)

print(f"zero = {gauche0}, zero = {ggauche0},1 = {gauche1}, 2 = {len(gauche2p)}")
print(gauche2p[0:10])


web = Grew_web()
print(web.url())
web.load_corpus(corpus)



import sys,os

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../")))  # Use local grew lib

from grewpy import Corpus, GRSDraft, Rule, Request, Commands, GRS, set_config

set_config("sud")

corpus = Corpus("examples/resources/test1.conllu")
graph = corpus[0]

print ("\n============================== TEST 1 ==============================")
print ("  Build a GRS from a file (examples/resources/test1.grs)")
grs = GRS("examples/resources/test1.grs")
print ("------------- print (grs) -------------")
print (grs)

print ("nb of output with strat s1 (should be 2) ---> ", end='')
print (len (grs.run(graph, strat="s1")))
print ("nb of output with strat s2 (should be 1) ---> ", end='')
print (len (grs.run(graph, strat="s2")))
print ("nb of output with strat s3 (should be 1) ---> ", end='')
print (len (grs.run(graph, strat="s3")))

print ("\n============================== TEST 2 ==============================")
print ("  Build a GRS from a string")
string_grs = """
rule det {
  pattern { N1[upos=DET]; N2[upos=NOUN]; N1 < N2 }
  without { N2 -> N1 }
  commands { add_edge N2 -[det]-> N1}
}

strat s1 { det }
strat s2 { Onf (det) }
strat s3 { Iter (det) }
"""

grs = GRS(string_grs)
print ("------------- print (grs) -------------")
print (grs)

print ("nb of output with strat s1 (should be 2) ---> ", end='')
print (len (grs.run(graph, strat="s1")))
print ("nb of output with strat s2 (should be 1) ---> ", end='')
print (len (grs.run(graph, strat="s2")))
print ("nb of output with strat s3 (should be 1) ---> ", end='')
print (len (grs.run(graph, strat="s3")))

print ("\n============================== TEST 3 ==============================")
print ("  Build a GRS with explicit values")

req_det_n = Request("N1[upos=DET]","N2[upos=NOUN]; N1 < N2").without("N2 -> N1")
add_det_cde = Commands("add_edge N2 -[det]-> N1")
R = Rule(req_det_n, add_det_cde)
grs = GRSDraft({"det":R,"s1":"det","s2":"Onf(det)","s3":"Iter(det)"})
print ("------------- print (grs) -------------")
print (grs)
print("-------- print corresponding abstract grs--------")
agrs = GRS(grs)
print(agrs)

print("----------- run grs with some strategies-----------")
print ("nb of output with strat s1 (should be 2) ---> ", end='')
print (len (agrs.run(graph, strat="s1")))
print ("nb of output with strat s2 (should be 1) ---> ", end='')
print (len (agrs.run(graph, strat="s2")))
print ("nb of output with strat s3 (should be 1) ---> ", end='')
print (len (agrs.run(graph, strat="s3")))

print("----------------test if grs can be saved-------------")
s = str(grs)
grs2 = GRSDraft(s)
print(grs2)

print("\n============================== TEST 4 ==============================")
print("  Visiting a GRS")
print("------------- print (rules of grs) -------------")
for d in grs.rules():
    print(f"rule name: {d}\n{grs[d]} \n")

exit(0)



from sklearn import tree
from sklearn.tree import DecisionTreeClassifier
import matplotlib.pyplot as plt
from sklearn.tree import plot_tree
import math
import numpy as np
import re
import argparse
import pickle

from rule_forgery import WorkingGRS, adjacent_rules, local_rules, refine_rules

# Use local grew lib
import os, sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../"))) 

from grewpy import Request, Rule, Commands, Command, Add_edge, GRSDraft, CorpusDraft, Package, Delete_feature
from grewpy import Corpus, GRS, set_config
from grewpy import grew_web
from grewpy.graph import Fs_edge

def add_rule_name(wgrs):
    for rn in wgrs:
        wgrs[rn].commands.append(Command(f'Y.rule="{rn}"'))

def clear_but_working(g):
    """
    delete non working edges within g
    """
    g.sucs = {n : [] for n in g.sucs}
    return (g)

def append_head(g):
    """
    each node in g get a head feature = 1
    """
    for n in g:
        g[n]['head']='1'
    return g

def basic_edges(g):
    """
    change edges {1:comp, 2:obl} to {1:comp}
    """
    def remove(e):
        return Fs_edge({'1': e['1']})
    for n in g.sucs:
        g.sucs[n] = tuple((m, remove(e)) for m, e in g.sucs[n])
    return g

def get_best_solution(corpus_gold, corpus_start, grs : GRS, strategy="main", verbose=0) -> CorpusDraft:
    """
    grs is a GRSDraft
    return the best solution using the grs with respect to the gold corpus
    the grs does not need to be confluent. We take the best solution with 
    respect to the f-score
    """
    def f_score(t):
        return t[0]/math.sqrt((t[0]+t[1])*(t[0]*t[2])+1e-20)

    solutions = grs.run(corpus_start)
    corpus = CorpusDraft(corpus_start)
    print(len(corpus_gold))
    i = 0
    for sid in corpus_gold:
        if i % (len(corpus_gold)//1) == 0:
            print(i)
        i += 1
        #gs = grs.run(corpus_start[sid], 'main')
        best_fscore = 0
        for g in solutions[sid]:
            fs = f_score(g.edge_diff_up_to(corpus_gold[sid]))
            if fs > best_fscore:
                best_fscore = fs
                corpus[sid] = g
    return corpus

def pack(s):
    if re.search("f[XYZ]", s):
        return 3
    if re.search("intermediate", s):
        return 2
    if "enhanced" in s:
        return 1
    return 0

def prepare_corpus(filename):
    corpus = Corpus(CorpusDraft(filename).apply(append_head))
    empty = Corpus(CorpusDraft(corpus).apply(clear_but_working))
    return corpus, empty

def append_delete_head(grs : GRSDraft):
    for rn in grs:
        grs[rn].commands.append(Delete_feature("Y", "head"))
    return grs

if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog='learner.py',
                                     description='Learn a grammar from sample files',
                                     epilog='Use connll files')

    parser.add_argument('train')
    parser.add_argument('-e', '--eval', default=None)
    parser.add_argument('--rules', default=None)
    parser.add_argument('-d','--nodetails',action="store_true")
    parser.add_argument('--verbose',default=0)
    parser.add_argument('-c', '--local_rules',action="store_true")
    parser.add_argument('-w','--web',action='store_true')
    args = parser.parse_args()
    if not args.eval:
        args.eval = args.train

    set_config("sud")
    param = {
        "base_threshold": 0.25,
        "valid_threshold": 0.90,
        "max_depth": 4,
        "min_samples_leaf": 5,
        "feat_value_size_limit": 10,
        "skip_features": ['xpos', 'upos', 'SpaceAfter', 'Shared', 'head'],
        "node_impurity": 0.2,
        "number_of_extra_leaves": 5, 
        "zipf_feature_criterion" : 0.95, 
        "min_occurrence_nb" : 10
    }
    corpus_gold, corpus_empty = prepare_corpus(args.train)
    if args.nodetails:
        corpus_gold = Corpus(CorpusDraft(corpus_gold).apply(basic_edges))
    
    A = corpus_gold.count(Request('X<Y;e:X->Y'))
    A += corpus_gold.count(Request('Y<X;e:X->Y'))
    print("---target----")
    print(f"""number of edges within corpus: {corpus_gold.count(Request('e: X -> Y'))}""")
    print(f"number of adjacent relations: {A}")

    R0,L0 = adjacent_rules(corpus_gold, param)
    print(len(L0))
    R0e = refine_rules(R0, corpus_gold, param)
    R0e = R0e.safe_rules()
    add_rule_name(R0e)
    R0e = append_delete_head(R0e)
    print(f"number of rules len(R0e) = {len(R0e)}")
    #turn R0e to a set of packages to speed up the process
    packages = {f'P{i}' : Package() for i in range(4)}
    for rn in R0e:
        packages[f'P{pack(rn)}'][rn] = R0e[rn]
    R00 = GRSDraft(packages)
    R00['main'] = "Seq(Onf(P0),Onf(P1),Onf(P2),Onf(P3))"
    if args.rules:
        f = open(args.rules, "w")
        f.write( str(R00) )
        f.close()

    G00 = GRS(R00)
    if args.web:
        web = grew_web.Grew_web()
        print(web.url())
        web.load_corpus(corpus_empty)
        web.load_grs(G00)

    currently_computed_corpus = get_best_solution(corpus_gold, corpus_empty, G00, args.verbose)
    print(currently_computed_corpus.edge_diff_up_to(corpus_gold))

    if args.local_rules:
        loc_rules,_ = local_rules(corpus_gold, param)
        Rloc = refine_rules(loc_rules, corpus_gold, param)
        Rloc = Rloc.safe_rules()
        Rloc = append_delete_head(Rloc)
        print(f"number of local rules: {len(Rloc)}")
        packages['P4'] = Package(Rloc)
        R00 = GRSDraft(packages)
        R00['main'] = "Seq(Onf(P0),Onf(P1),Onf(P2),Onf(P3),Onf(P4))"
        currently_computed_corpus = get_best_solution(corpus_gold, currently_computed_corpus, G00, strategy="Onf(P4)")
        print(currently_computed_corpus.edge_diff_up_to(corpus_gold))
        if args.rules:
            f = open(args.rules, "w")
            f.write(str(R00))
            f.close()
        print('done')

"""

def edge_is(e,f):
    '''
    tells whether e is an f
    '''
    for k in e:
        if k not in f or e[k] not in f[k]:
            return False
    return True

def corpus_diff(c1,c2):
    c,l,r = 0,0,0
    for sid in c1:
        men = {m: (e, n) for n, e, m in c1[sid].triples()}
        menp = {m : (e,n) for n,e, m in c2[sid].triples()}
        S1 = set(men.keys())
        S2 = set(menp.keys())
        l += len(S1 - S2)
        r += len(S2 - S1)
        for m in S1 & S2:
            e,n = men[m]
            ep,np = menp[m]
            if n != np or not edge_is(ep,e):
                l,r = l+1,r+1
            else:
                c += 1
    precision = c / (c + l+1e-10)
    recall = c / (c + r+1e-10)
    f_measure = 2*precision*recall / (precision+recall+1e-10)
    return {
        "common": c,
        "left": l,
        "right": r,
        "precision": round(precision, 3),
        "recall": round(recall, 3),
        "f_measure": round(f_measure, 3),
    }
"""



from .grs import Request
class Sketch:
    def __init__(self, P, cluster_criterion, avec, without, target):
        """
        sketches are defined by a pattern P, and parameters (features, edge features, etc)
        sketch name serves for rule naming
        Nodes X and Y have special status. 
        In the following, we search for 'e : X -> Y'
        Thus, in the pattern P, X, Y, e must be used with that fact in mind.
        cluster criterion is typically ["X.upos", "Y.upos"], but can be enlarged  
        """
        self.P = P
        self.cluster_criterion = cluster_criterion
        self.avec = avec
        self.without = without
        self.target = target

    def cluster(self, corpus):
        def _none_(n):
            return '' if '__none__' in n else n[0]
        obs2 = corpus.count(self.P, self.cluster_criterion, ["X->Y"], True)
        obs3 = dict()
        for k1, o1 in obs2.obs.items():
            if len(o1) == 1 and ('__none__',) in o1:
                ...
            else:
                obs3[k1] = {_none_(k2) : v for k2,v in o1.items()}
        return obs3



class Observation:
    """
    maps a tuple of criteria to a dict mapping edge -> nb of observation
    """
    @staticmethod
    def flatten(obs, crit):
        def _flatten(obs, crit, L): #flatten a dict according to crit
            if not obs:
                return
            if not crit:
                yield (L, obs)
            else:
                for k in obs:
                    yield from _flatten(obs[k], crit[1:], L + (k,))
        return {L : v for L, v in _flatten(obs, crit, tuple())}

    def __init__(self, **kwargs):
        if "obs" in kwargs:
            if "parameter" in kwargs:
                intermediate = Observation.flatten(kwargs["obs"], kwargs["parameter"])
            if "keys" in kwargs:
                self.obs = {L : Observation.flatten(V, kwargs["keys"]) for L,V in intermediate.items()}
            else:
                self.obs = intermediate
        else:
            self.obs = dict()


    def __ior__(self, other):
        for parameter_keys in other.obs:
            if parameter_keys not in self.obs:
                self.obs[parameter_keys] = dict()
            for observation_keys in other.obs[parameter_keys]:
                if observation_keys not in self.obs[parameter_keys]:
                    self.obs[parameter_keys][observation_keys] = 0
                self.obs[parameter_keys][observation_keys] += other.obs[parameter_keys][observation_keys]
        return self

    def __iter__(self):
        return iter(self.obs)

    def __getitem__(self, k):
        return self.obs[k]

    def __setitem__(self, k, v):
        self.obs.__setitem__(k,v)

    def __bool__(self):
        return bool(self.obs)

    def anomaly(self, L,  threshold):
        """
        L is a key within self
        return for L an edge and its occurrence evaluation 
        and number of total occurrences if beyond base_threshold
        """
        s = sum(self.obs[L].values())
        for x, v in self.obs[L].items():
            if v > threshold * s and x:
                return (x,v,s)
        return None, None, None

  
    def zipf(observation, n, k, width, ratio):
        """
        return the list of width best features
        if they are beyond ratio
        """
        if len(observation[(n, k)]) < 1:
            return []  # no values or 1 is not sufficient
        values = list(observation[(n, k)].keys())
        values.sort(reverse=True)
        occs = sum([observation[(n, k)][v] for v in values])
        zoccs = sum([observation[(n, k)][v] for v in values[0:width]])
        if zoccs/occs > ratio:
            return values[:width]
        return []



import sys, os, json

sys.path.insert(0, os.path.abspath(os.path.join( os.path.dirname(__file__), "../"))) # Use local grew lib

# from grewpy import Graph, CorpusDraft, Request, Corpus, GRS, request_counter
from grewpy import Corpus, GRSDraft, Rule, Request, Commands, GRS

from grewpy import set_config
set_config("sud")

pud_file = "examples/resources/fr_pud-sud-test.conllu"
pud = Corpus(pud_file)
print(f"nb of graph in {pud_file} = {len(pud)}")


# g = pud[9]
# print(g)
# print (g.edge('8', '4'))

r = Request('e: N -[1=subj]-> M;')
print (pud.count(r))

x = Request('e: N -[subj]-> M;')
print (pud.count(x))

exit (0)

r = Request('e: N -> M; M[form="a"]')
print (r)

matchings = pud.search(r, ['e.label'])
print (json.dumps(matchings, indent=2)) 
{'1': 'nsubj'}
print ("------------------------------------------")

grs = GRS ("""
rule r { pattern { N[upos=NOUN] } commands { N.upos=N} } 
""")


# grs = GRS ("""
# rule r {
#   pattern { 
#     N -[1=nsubj]-> M
#   } 
#   without { N -[xxx]-> M }
#   commands { add_edge N -[xxx]-> M} } 
# """)

o = grs.run(pud, 'r')

print (type(o))

for sent_id, graph_list in o.items():
  print (f"{sent_id} --> {len(graph_list)}")

print ("======================================")

g = pud[9]
print (type(g))
g2 = grs.apply(g, 'Onf(r)')
print (type(o))

print (g2.edge('8', '4'))
# for sent_id get_sent_ids(o):
#   print (sent_id)
#   # print (g.to_json())



"""
Grew module : anything you want to talk about graph rewriting systems
Graphs are represented by dictionaries (nodes, edges, meta informations)
GRS can be build via files, explicit constructions or even strings. See doc of GREW
"""
import typing
import json

from . import network

''' Library tools '''

JSON = typing.Any #dict[str,typing.Any] | list[typing.Any] | str | int

class GrewError(Exception):
    """A wrapper for grew-related errors"""

    def __init__(self, message):
        self.value = message
    def __str__(self):
        if isinstance(self.value, dict):
            return ("\n".join (("", "-"*80, json.dumps(self.value, indent=2), "-"*80)))
        else:
            return ("\n".join (("", "="*80, str (self.value), "="*80)))

GrewError.__doc__ = "A wrapper for grew-related errors"

def set_config(data):
    """
    Change the configuration used in the next exchanges
    See https://grew.fr/doc/graph/#edges for details about config
    """
    return network.send_and_receive({"command": "set_config", "config": data})

def request_counter():
    return network.request_counter



#===========================================================================
# used for launching grew web on a given corpus
import tempfile
import requests
import webbrowser
import http
import json
from .grew import GrewError
from grewpy import GRSDraft


local = False
if local:
    grew_web_back = "http://localhost:8080"
    grew_web_front = "http://localhost:8888/grew_web"
else:
    grew_web_back = "https://gwb.grew.fr"
    grew_web_front = "https://web.grew.fr"


def _post_request(service, resp):
    if resp.status_code >= 300:
        raise GrewError({
            "Error": "HTTP",
            "service": service,
            "status_code": resp.status_code,
            "message": http.client.responses[resp.status_code]
        })
    data = json.loads(resp.text)
    if data["status"] == "OK":
        return data["data"]
    if data["status"] == "ERROR":
        raise GrewError({"grew_web": service, "message": data["message"]})
    raise GrewError({"UNEXPECTED": service, "message": data})


class Grew_web:
    """
    web-connection to grew
    """
    def __init__(self):
        self.session_id = _post_request(
            "connect", requests.post(f"{grew_web_back}/connect"))

    def load_grs(self,grs):
        json_grs = grs.json()
        if self.session_id == "":
            raise GrewError({"grew_web": "not connected"})
        with tempfile.NamedTemporaryFile(mode="a+", delete=True, suffix=".grs") as f:
            f.write(json.dumps(json_grs))
            f.seek(0)  # ready to be read
            r = requests.post(f"{grew_web_back}/upload_grs",
                          data={"session_id": self.session_id},
                          files={"json_file": f}
                          )
            try:
                _post_request("upload_grs", r)
            except GrewError:
                with open("error.grs","w") as g:
                    g.write(str(GRSDraft(grs)))
                


    def load_corpus(self,   corpus):
        conll = corpus.to_conll()
        if self.session_id == "":
            raise GrewError({"grew_web": "not connected"})
        with tempfile.NamedTemporaryFile(mode="a+", delete=True, suffix=".conllu") as f:
            f.write(conll)
            f.seek(0)  # ready to be read
            r = requests.post(f"{grew_web_back}/upload_corpus",
                          data={"session_id": self.session_id},
                          files={"file": f}
                          )
            _post_request("upload_corpus", r)


    def url(self):
        if self.session_id == "":
            raise GrewError({"grew_web": "not connected"})
        return f"{grew_web_front}?session_id={self.session_id}"

    def open(self):
        webbrowser.open (self.url())



import json
import os.path

from . import network
from .grew import JSON
from grewpy.graph import Graph
from .corpus import Corpus, CorpusDraft, GrewError

class RequestItem():
    def __init__(self,sort : str,*L):
        """
        sort in {"global", "pattern", "without", "with"}
        L is a list of
         - ";" separated clauses or
         - a list of items
         - they will be concatenated
        """
        self.sort = sort
        self.items = tuple()
        for elt in L:
            if isinstance(elt,str):
                self.items += tuple(c.strip() for c in elt.split(";") if c.strip())
            else:
                self.items += tuple(elt)

    def json_data(self):
        return {self.sort : list(self.items)}

    @classmethod
    def from_json(cls, json_data : JSON) :
        k = list(json_data)[0]
        v = json_data[k]
        return cls(k,*v)

    def __str__(self):
        its = ";".join([str(x) for x in self.items])
        return f"{self.sort} {{{its}}}"

    def __repr__(self):
        return f"{self.sort} {{{ ';'.join([str(x) for x in self.items]) }}}"

class Request():
    """
    lists of ClauseList
    """
    def __init__(self, *L):
        """
        L is either a list of
         - RequestItem or
         - (pattern) string or a
         - Request (for copies)
        """
        elts = tuple()
        for e in L:
            if isinstance(e,str):
                elts += (RequestItem("pattern", e),)
            elif isinstance(e,RequestItem):
                elts += (e,)
            elif isinstance(e,Request):
                elts += e.items
            elif isinstance(e,tuple): #supposed to be a list of ClauseList
                elts += e
            else:
                try:
                    #suppose it is a generator
                    for x in e:
                        elts += (x,)
                except:
                    raise ValueError(f"{e} cannot be used to build a Request")
        self.items = elts

    def without(self, *L):
        self.items += tuple(RequestItem("without", e) for e in L)
        return self
    def with_(self, *L):
        self.items += tuple(RequestItem("with", e) for e in L)
        return self


    def with_(self, *L):
        self.items += tuple(RequestItem("with", e) for e in L)
        return self

    @classmethod
    def from_json(cls,json_data):
        elts = [RequestItem.from_json(c) for c in json_data]
        return cls(*elts)

    def json_data(self):
        return [x.json_data() for x in self.items]

    def __str__(self):
        return "\n".join([str(e) for e in self.items])

    def __iter__(self):
        return iter(self.items)

    def pattern(self):
        """
        return the pattern of self as a tuple
        """
        return Request(p for p in self.items if p.sort == "pattern")

    def append(self, *L):
        """
        Append a new ClauseList to the Request
        L is given either as a pair (s,t) with "s \in {'pattern','without','meta'} and t : str 
        or L[0] is a ClauseList 
        """
        if len(L) == 2:
            self.items = self.items + (RequestItem(L[0], L[1]),)
        elif len(L) == 1:
            self.items = self.items + L
        else:
            raise ValueError(f"cannot build a clause list with {L}")


class Command:
    def __init__(self,s):
        """
        self.item = str representation of the command
        """
        self.item = s

    def json_data(self):
        return self.item
    
    def __str__(self):
        return str(self.item)

    def safe(self):
        """
        return a clause list for a safe request
        """
        raise NotImplementedError ("not yet implemented")

class Add_edge(Command):
    def __init__(self,X,e,Y):
        if isinstance(e, dict):
            s = ",".join(f"{k}={v}" for k,v in e.items())
        else:
            s = str(e)
        super().__init__(f"add_edge {X}-[{s}]->{Y}")
        self.X, self.e, self.Y = X, e, Y

    def safe(self):           
        return RequestItem("without",self.item.replace("add_edge",""))
    
    def __repr__(self):
        return str(self)

class Delete_edge(Command):
    def __init__(self, X, e, Y):
        super().__init__(f"del_edge {X}-[{e}]->{Y}")
        self.X, self.e, self.Y = X, e, Y

    def safe(self):           
        return RequestItem("pattern", self.item.replace("add_edge", ""))
    
class Delete_feature(Command):
    def __init__(self, X, f):
        super().__init__(f"del_feat {X}.{f}")
        self.X = X
        self.f = f


class Commands(list):
    def __init__(self, *L):
        super().__init__()
        for elt in L:
            if isinstance(elt,str):
                self += [t.strip() for t in elt.split(";") if t.strip()]
            elif isinstance(elt,list):
                self += elt
            elif isinstance(elt, Command):
                self.append(elt)

    def __str__(self):
        c = ";".join([str(x) for x in self])
        return f"commands {{{c}}}"

    @classmethod
    def from_json(cls, json_data):
        return cls(*json_data)
    
    def json_data(self):
        return [x if isinstance(x,str) else x.json_data() for x in self]

class Rule():
    def __init__(self, request : Request, cmd_list : Commands, lexicons =None):
        self.request = request
        self.commands = cmd_list
        self.lexicons = lexicons if lexicons else dict()

    def json_data(self):
        p = self.request.json_data()
        c = self.commands.json_data()
        return {"request" : p, "commands" : c, "lexicons" : json.dumps(self.lexicons)}

    def __str__(self):
        return f"{str(self.request)}\n{str(self.commands)}"

    def __repr__(self):
        return f"{str(self.request)}\n{str(self.commands)}"

    @classmethod
    def from_json(cls,json_data):
        # print(json_data)
        reqs = Request.from_json(json_data["request"])
        cmds = Commands.from_json(json_data["commands"])
        return cls(reqs,cmds)

class Package(dict):
    """
    dict mapping names to rule/package/strategies"""

    @classmethod
    def from_json(cls, json_data):
        res = Package._from_json(json_data)
        return cls(res)

    def _from_json(json_data):
        res = dict()
        for k,v in json_data.items():
            if isinstance(v,str):
                res[k] = v
            elif "decls" in v: #it is a package
                res[k] = Package.from_json(v["decls"])
            else:
                res[k] = Rule.from_json(v)
        return res

    def json_data(self):
        elts = dict()
        for k,v in self.items():
            elts[k] = v if isinstance(v,str) else v.json_data()
        return {"decls" : elts}

    def __str__(self):
        res = [f"strat {k} {{{self[k]}}}" for k in self.strategies()] +\
            [f"package {k} {{{str(self[k])}}}" for k in self.packages()] +\
            [f"rule {k} {{{str(self[k])}}}" for k in self.rules()]
        return "\n".join(res)

    def rules(self):
        return filter(lambda x: isinstance(self[x], Rule), self.__iter__())

    def packages(self):
        return filter(lambda x: isinstance(self[x], Package), self.__iter__())

    def strategies(self):
        return filter(lambda x: isinstance(self[x], str), self.__iter__())


class GRSDraft(Package):
    """
    A GRSDraft is a structure that gives access to the internals of a 
    Graph Rewriting System: packages, rules, patterns, strategies, etc
    It cannot be used to perform rewriting, for that, use a GRS
    """

    def __init__(self,args=None):
        """Load a grs stored in a file
        :param data: either a file name or a Grew string representation of a grs
        :or kwargs contains explicitly the parts of the grs
        :return: an integer index for latter reference to the grs
        :raise an error if the file was not correctly loaded
        """
        if isinstance(args,str):
            agrs = GRS(args)
            json_data = agrs.json()
            res = Package._from_json(json_data["decls"])
            super().__init__(res)
        elif isinstance(args, dict):
            super().__init__( args )
        elif args == None:
            super().__init__()

    def __str__(self):
        return super().__str__()

    def rules(self):
        return (rule_name for rule_name in self if isinstance(self[rule_name],Rule))

    def safe_rules(self):
        """
        create a new grs with application of safe to each rule. 
        self.rules() are supposed to contain only Commands of length 1 that support safe method
        """
        grs = GRSDraft()
        for rule_name in self.rules():
            rule = self[rule_name]
            cde = rule.commands[0]
            safe_request = Request(rule.request)
            safe_request.append(cde.safe())
            safe_rule = Rule(safe_request, rule.commands, rule.lexicons)
            grs[rule_name] = safe_rule
        return grs

    def onf(self, strat_name="main"):
        self[strat_name] = f'Onf(Alt({",".join(self.rules())}))'
        return self
    
    def save(self, filename):
        with open(filename, "w") as f:
            f.write(str(self))
      
class GRS:
    """
    An abstract GRS. Offers the possibility to apply rewriting.
    The object is abstract and cannot be changed. 
    For that, use a GRSDraft
    """

    def __init__(self, args):
        """Load a grs stored in a file
        :param data: either a file name or a Grew string representation of a grs
        :or kwargs contains explicitly the parts of the grs
        :return: an integer index for latter reference to the grs
        :raise an error if the file was not correctly loaded
        """
        if isinstance(args, str):
            if os.path.isfile(args):
                req = {"command": "load_grs", "file": args}
            else:
                req = {"command": "load_grs", "str": args}
        elif isinstance(args, GRSDraft):
            req = {"command": "load_grs", "json": args.json_data()}
        elif isinstance(args, dict):
            """
            suppose it is a GRS style
            """
            try:
                grs = GRSDraft(args)
                req = {"command": "load_grs", "json": grs.json_data()}
            except GrewError as e:
                raise ValueError(f"cannot build a grs with {args}\n {e.message}")
        else:
            raise ValueError(f"cannot build a grs with {args}")
    
        reply = network.send_and_receive(req)
        self.id = reply["index"]    

    def json(self):
        req = {"command": "json_grs", "grs_index": self.id}
        return network.send_and_receive(req)

    def __str__(self):
        return f"GRS({self.id})"

    def run(self, data, strat="main"):
        """
        run a Grs on a graph
        :param data: a graph or an AbstractCorpus
        :param strat: the strategy (by default "main")
        :return: a dictionary mapping sid to the list of rewritten graphs
        """
        if isinstance(data, Graph):
            req = {
                "command": "grs_run_graph",
                "graph": json.dumps(data.json_data()),
                "grs_index": self.id,
                "strat": strat
            }
            reply = network.send_and_receive(req)
            return [Graph.from_json(s) for s in reply]
        elif isinstance(data, Corpus):
            req = {
                "command": "grs_run_corpus",
                "corpus_index": data.get_id(),
                "grs_index": self.id,
                "strat": strat
            }
            reply = network.send_and_receive(req)
            return {sid: [Graph(s) for s in L] for sid, L in reply.items() } 
        elif isinstance(data, CorpusDraft):
            return {sid: self.run(g) for sid,g in data.items() } 

    def apply(self, data, strat="main", abstract=True):
        """
        run a Grs on a graph or corpus
        :param grs_data: a graph rewriting system or a Grew string representation of a grs
        :param G: the graph, either a str (in grew format) or a dict
        :param strat: the strategy (by default "main")
        :return: the rewritten graph and an error if there is not exaclty one output graph
        """
        if isinstance(data, Graph):
            req = {
                "command": "grs_apply_graph",
                "graph": json.dumps(data.json_data()),
                "grs_index": self.id,
                "strat": strat
            }
            reply = network.send_and_receive(req)
            return Graph(reply)
        elif isinstance(data, Corpus):
            req = {
                "command": "grs_apply_corpus",
                "corpus_index": data.get_id(),
                "grs_index": self.id,
                "strat": strat
            } # return None because inplace
            network.send_and_receive(req)
            return data if abstract else CorpusDraft (data)
        elif isinstance(data, CorpusDraft):
            acorpus = Corpus(data)
            self.apply(acorpus, strat, abstract)

    



from sklearn import tree
from sklearn.tree import DecisionTreeClassifier
import math
import numpy as np
# Use local grew lib
import os
import sys
sys.path.insert(0, os.path.abspath(
    os.path.join(os.path.dirname(__file__), "../")))

from grewpy.sketch import Sketch
from grewpy.observation import Observation
from grewpy import Corpus, GRS, set_config
from grewpy import Request, Rule, Commands, Add_edge, GRSDraft, CorpusDraft


def feature_value_occurences(matchings, corpus):
    """
    given a matchings corresponding to some request on the corpus,
    return a dict mapping (n,feature) =>(values)=>occurrences to its occurence number in matchings
    within the corpus. n : node name, feature like 'Gender', values like 'Fem'
    """
    observation = Observation()
    for m in matchings:
        graph = corpus[m["sent_id"]]
        nodes = m['matching']['nodes']
        for n in nodes:
            N = graph[nodes[n]]  # feature structure of N
            for k, v in N.items():
                if (n, k) not in observation:
                    observation[(n, k)] = dict()
                observation[(n, k)][v] = observation[(n, k)].get(v, 0)+1
    return observation

def feature_values_for_decision(matchings, corpus, param, nodes):
    """
    restrict feature values to those useful for a decision tree
    """
    observation = feature_value_occurences(matchings, corpus)
    features = dict()
    for (n,k) in observation:
        if n in nodes and k not in param["skip_features"]:
            for v in observation.zipf(n, k, param["feat_value_size_limit"], param["zipf_feature_criterion"]):
                features[(n,k,v)] = observation[(n,k)][v]
    return features

class Classifier():
    def __init__(self, matchings, corpus, param):
        self.fpat = list(feature_values_for_decision(matchings, corpus, param, ['X', 'Y']).keys())
        # get the list of all feature values
        self.pos = {self.fpat[i]: i for i in range(len(self.fpat))}
        X, y1, y = list(), dict(), list()
        # X: set of input values, as a list of (0,1)
        # y: set of output values, an index associated to the edge e
        # the absence of an edge has its own index
        # y1: the mapping between an edge e to some index
        for m in matchings:
            # each matching will lead to an obs which is a list of 0/1 values
            graph = corpus[m["sent_id"]]
            nodes = m['matching']['nodes']
            obs = [0]*len(self.pos)
            for n in nodes:
                feat = graph[nodes[n]]
                for k, v in feat.items():
                    if (n, k, v) in self.pos:
                        obs[self.pos[(n, k, v)]] = 1
            es = graph.edges(nodes['X'], nodes['Y'])
            if len(es) > 1:
                print("mmmmhh that should not happen")
            elif len(es) <= 1:
                e = es.pop() if es else None
                if e not in y1:
                    y1[e] = len(y1)
                y.append(y1[e])
                X.append(obs)
        if not X or not len(X[0]) or max(y) == 0:
            self.clf = None
        else:
            self.clf = DecisionTreeClassifier(max_depth=param["max_depth"],
                                              max_leaf_nodes=max(
                y)+param["number_of_extra_leaves"],
                min_samples_leaf=param["min_samples_leaf"],
                criterion="gini")
            self.clf.fit(X, y)
            self.y1 = {y1[i]: i for i in y1}

    def branches(self, pos, current, acc, threshold):
        tree = self.clf.tree_
        if tree.feature[pos] >= 0:
            if tree.impurity[pos] < threshold:
                acc[pos] = current
                return
            # there is a feature
            if tree.children_left[pos] >= 0:
                # there is a child
                left = current + ((tree.feature[pos], 1),)
                self.branches(tree.children_left[pos],
                              left, acc, threshold)
            if tree.children_right[pos] >= 0:
                right = current + ((tree.feature[pos], 0),)
                self.branches(tree.children_right[pos],
                              right, acc, threshold)
            return
        else:
            if tree.impurity[pos] < threshold:
                acc[pos] = current

    def find_classes(clf, param):
        """
    given a decision tree, extract "interesting" branches
    the output is a dict mapping the node_index to its branch
    a branch is the list of intermediate constraints = (7,1,8,0,...)
    that is feature value 7 has without clause whereas feature 8 is positive 
        """
        acc = dict()
        clf.branches(0, tuple(), acc, param["node_impurity"])
        return acc



"""
Grew python library
All you need to use grew 
See grew online documentation for global informations
"""
from .corpus import CorpusDraft, Corpus
from .grs import Request, GRSDraft, Package, Rule, Commands, GRS, Add_edge, Delete_edge, Delete_feature, Command
from .graph import Graph
from .grew import set_config, request_counter

from .network import init
init()


"""
Grew module : anything you want to talk about graphs
Graphs are represented either by a dict (called dict-graph),
or by an str (str-graph).
"""
import os.path
import glob

import sys
import tempfile
import json
import typing
import numpy as np

from .network import send_and_receive
from .graph import Graph
from .grew import GrewError
from .observation import Observation
from . import network

from .matchings import Matchings

class AbstractCorpus():
    def edge_diff(self, other, edge_criterion=lambda e: True):
        """
        given two corpora, outputs the number of common edges, only left ones and only right ones.
        It also outputs precision, recall and f-measure.
        """
        (common, left, right) = np.sum(
            [self[sid].edge_diff(other[sid],edge_criterion) for sid in self], axis=0)
        precision = common / (common + left+1e-10)
        recall = common / (common + right+1e-10)
        f_measure = 2*precision*recall / (precision+recall+1e-10)
        return {
            "common": common,
            "left": left,
            "right": right,
            "precision": round(precision, 3),
            "recall": round(recall, 3),
            "f_measure": round(f_measure, 3),
        }

    def edge_diff_up_to(self, other, edge_transform=lambda e: e):
        (common, left, right) = np.sum(
            [self[sid].edge_diff_up_to(other[sid], edge_transform) for sid in self], axis=0)
        precision = common / (common + left)
        recall = common / (common + right)
        f_measure = 2*precision*recall / (precision+recall)
        return {
            "common": common,
            "left": left,
            "right": right,
            "precision": round(precision, 3),
            "recall": round(recall, 3),
            "f_measure": round(f_measure, 3),
        }

class CorpusDraft(AbstractCorpus,dict):
    """
    the draft is composed of 
      - self, a dict mapping sentence_id to graphs
      - self._sent_ids, a list that specifies the sentence order
    """
    def __init__(self,data=None):
        """Load a corpus from a file of a string
        :param data: a file, a list of files or a CoNLL string representation of a corpus
        :return: an integer index for latter reference to the corpus
        :raise an error if the files was not correctly loaded
        """
        if isinstance(data, CorpusDraft):
            T = {sid: Graph(data[sid]) for sid in data}
            super().__init__(T)
        elif isinstance(data, dict):
            super().__init__(data)
        elif data == None:
            super().__init__()
        else:
            acorpus = data if isinstance(data, Corpus) else Corpus(data)
            self._sent_ids = acorpus.get_sent_ids() #specifies the sentences order
            super().__init__(acorpus.get_all())

    def __getitem__(self, data):
        """
        Search for [data] in previously loaded corpus
        :param data: a sent_id (type string) or a position (type int)
        :param corpus_index: an integer given by the [corpus] function
        :return: a graph
        """
        if isinstance(data, str):
            return super().__getitem__(data)
        if isinstance(data, int):
            return self[self._sent_ids[data]]
        if isinstance(data, slice):
            return [self[sid] for sid in self._sent_ids[data]]

    def apply(self, fun):
        """
        Apply fun to all graphs, return the new Corpus
        """
        return CorpusDraft({sid : fun(self[sid]) for sid in self})

    def to_conll(self):
        return Corpus(self).to_conll()

class Corpus(AbstractCorpus):
    def __init__(self, data):
        """An abstract corpus
        :param data: a file, a list of files or a CoNLL string representation of a corpus
        :return: an integer index for latter reference to the corpus
        :raise an error if the files was not correctly loaded
        """
        if isinstance(data, list):
            if data and isinstance(data[0], Graph):
                graphs = {f'{i}' : data[i].json_data() for i in range(len(data))}
                req = {"command": "corpus_from_dict", "graphs": graphs}
            else:
                #supposed to be a list of files
                req = {"command": "corpus_load", "files": data}
            reply = network.send_and_receive(req)
        elif isinstance(data, dict):
            req = {"command": "corpus_from_dict", "graphs": {
                sent_id: graph.json_data() for (sent_id, graph) in data.items()}}
            reply = network.send_and_receive(req)
        elif os.path.isdir(data):
            # load of connlu files of the directory
            file_list = glob.glob(f"{data}/*.conllu") + glob.glob(f"{data}/*.conll") + glob.glob(f"{data}/*.cupt")
            req = {"command": "corpus_load", "files": file_list}
            reply = network.send_and_receive(req)
        elif os.path.isfile(data):
            req = {"command": "corpus_load", "files": [data]}
            reply = network.send_and_receive(req)
        else:
            with tempfile.NamedTemporaryFile(mode="w", delete=True, suffix=".conll") as f:
                f.write(data)
                f.flush()  # to be read by others
                req = {"command": "corpus_load", "files": [f.name]}
                try:
                    reply = network.send_and_receive(req)
                except GrewError:
                    raise GrewError(data)
        self._length = reply["length"]
        self._id = reply["index"]

    def get_sent_ids(self):
        """
        return the list of sentence ids
        """
        req = {"command": "corpus_sent_ids", "corpus_index": self._id}
        return network.send_and_receive(req)

    def get_id(self):
        """
        return the id of the corpus
        """
        return self._id

    def clean(self):
        """
        clean the corpus (remove from the backend memory)
        """
        req = {"command": "corpus_clean", "corpus_index": self._id}
        return network.send_and_receive(req)

    def get(self, sent_id):
        """
        return a graph corresponding to the sentence id sent_id
        """
        req = {"command": "corpus_get",
                   "corpus_index": self._id, "sent_id": sent_id}
        return (Graph.from_json(network.send_and_receive(req)))
    
    def __getitem__(self, data):
        """
        return a graph corresponding to data, either
          - a sentence id,
          - an index in the sentence id array
          - a slice
        """
        if isinstance(data, str):
            return self.get(data)
        if isinstance(data, int):
            sids = self.get_sent_ids()
            return self.get(sids[data])
        if isinstance(data, slice):
            sids = self.get_sent_ids()
            return [self[sid] for sid in sids[data]]


    def get_all(self):
        """
        return a dictionary mapping sentence ids to graphs
        """
        dico = network.send_and_receive({"command": "corpus_get_all", "corpus_index": self._id})
        return {sid: Graph.from_json(json_data) for (sid,json_data) in dico.items() }


    def search(self, request, clustering_parameter=[], clustering_keys=[],flat=None):
        """
        Search for [request] into [corpus_index]

        Parameters:
        request (Request): a request
        corpus_index: an integer given by the [corpus] function

        Returns:
        list: the list of matching of [request] into the corpus
        """
        res = network.send_and_receive({
            "command": "corpus_search",
            "corpus_index": self._id,
            "request": request.json_data(),
            "clustering_keys": clustering_parameter + clustering_keys
        })
        if flat == "matchings":
            return Matchings(res, self)
        elif flat == "observations" and clustering_parameter or clustering_keys:
            return Observation(res, clustering_parameter, clustering_keys)
        return res

    def count(self, request, clustering_parameter=[], clustering_keys=[], flat=False):
        """
        Count for [request] into [corpus_index]
        :param request: a string request
        :param corpus_index: an integer given by the [corpus] function
        :return: the number of matching of [request] into the corpus
        """
        res = network.send_and_receive({
            "command": "corpus_count",
            "corpus_index": self._id,
            "request": request.json_data(),
            "clustering_keys": clustering_parameter + clustering_keys,
        })
        if not flat:
            return res
        if clustering_parameter or clustering_keys:
            return Observation(obs=res,parameter=clustering_parameter, keys=clustering_keys)
        return res

    def __len__(self):
        return self._length

    def __iter__(self):
        return iter(self.get_sent_ids())

    def to_conll(self):
        """
        return a CoNLL string for the current corpus
        """
        reply = network.send_and_receive({
            "command": "corpus_to_conll",
            "corpus_index": self._id
        })
        return reply



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='grewpy',
    version='0.4.1',
    packages=['grewpy','examples'],
    license='LICENSE/Licence_CeCILL_V2-en.txt',
    description="A binding to the Grew software",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://grew.fr",
    author="bguil",
    author_email="Bruno.Guillaume@loria.fr"
)



from .observation import Observation

class Matching():
    """
    a class that represent a matching
    """
    def __init__(self, json_data, graph):
        self.nodes = json_data["nodes"]
        self.edges = json_data["edges"]
        self.graph = graph

    def feature_values(self, arg=None, flat=False):
        """
        return the list of feature values of arg
        arg = None = all nodes, an explicit list of nodes, a unique node
        if flat, all values are flatten, otherwise, computation is done for each node
        """
        nl = self.nodes if not arg else arg if isinstance(arg, list) else [arg]
        if flat:
            """
            not yet implemented
            """
            ...
        else:
            observation = Observation()
            for n in nl:
                n_in_g = self.nodes[n]
                for k,v in self.graph[n_in_g].items():
                    if (n, k) not in observation:
                        observation[(n, k)] = dict()
                    observation[(n, k)][v] = observation[(n, k)].get(v, 0)+1
            return observation


class Matchings(dict):
    """
    matching sentence ids to list of matchings
    """
    def __init__(self, json_data, corpus):
        super().__init__()
        for line in json_data:
            sid = line["sent_id"]
            if sid not in self:
                self[sid] = []
            self[sid].append(Matching( line["matching"], 
            corpus[sid]))

    def feature_values(self, arg=None, flat=False):
        observation = Observation()
        for ms in self.values():
            for m in ms: #for each matching of the sentence id
                observation |= m.feature_values(arg,flat)
        return observation

    

